#include <arpa/inet.h>
#include <iostream>
#include <netinet/in.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <ostream>
#include <sys/socket.h>
#include <unistd.h>

using std::cout, std::cin, std::endl, std::string;

/*
 * Remember to build with the flags "-L/usr/lib -lssl -lcrypto"
 */

/**
 * Gets the latest error from OpenSSL and returns it as an string.
 * This makes it easier to log any errors that OpenSSL throws.
 */
string getCtxError() {

    // Get the latest error code
    unsigned long errCode = ERR_get_error();

    // Define a buffer of 128 bytes to hold the error message
    char errBuffer[128];

    // Get a human readable description of the error from OpenSSL and write it
    // to the buffer
    ERR_error_string_n(errCode, errBuffer, sizeof(errBuffer));
    return errBuffer;
}

int main() {

    // Define a structure to contains the content of a message and the user who
    // sent it
    struct Message {
        string content;
        string username;

        string toString() {
            return "User '" + username + "' wrote:\n" + content;
        }
    };

    // Initializes the SSL library
    if (OpenSSL_version_num() < 0x10100000L) {
        // Old version, deprecated as of version 1.1.0
        SSL_library_init();

        // Loads human readable error strings
        // Automatically initialized in newer versions
        SSL_load_error_strings();
    } else {
        // New version
        // Initializes both SSL and crypto and is generally better
        OPENSSL_init_ssl(0, nullptr);
    }

    // Create new SSL context with client side connections and save it to a
    // pointer
    SSL_CTX *ctx = SSL_CTX_new(TLS_client_method());
    if (!ctx) {
        cout << "Creation of SSL context failed: " << getCtxError() << endl;
        return -1;
    }

    // Load the server's certificate into context
    // Returns 1 on success, otherwise check error stack
    if (SSL_CTX_load_verify_locations(ctx, "./ssl/server.crt", nullptr) <= 0) {
        cout << "Certificate load failed: " << getCtxError() << endl;
        return -1;
    }

    // Create SSL object
    SSL *ssl = SSL_new(ctx);
    if (!ssl) {
        cout << "Failed to create SSL object" << endl;
    }

    // Initializes a new TCP socket in the IPv4 family
    // 0 tells the OS to automatically pick the correct protocol (TCP)
    int clientSocket = socket(AF_INET, SOCK_STREAM, 0);
    if (clientSocket == -1) {
        cout << "Client socket failed" << endl;
        return clientSocket;
    }

    // Attach our SSL object to the socket
    SSL_set_fd(ssl, clientSocket);

    sockaddr_in serverAdress;

    // Specify that we use the IPv4 family
    serverAdress.sin_family = AF_INET;

    // Define the port the client should connect to
    serverAdress.sin_port = htons(8080);

    // Define the IP the client should connect to
    serverAdress.sin_addr.s_addr = inet_addr("127.0.0.50");

    // Connect the client to the server
    // serverAdress is cast to a sockaddr
    int connectStatus = connect(clientSocket, (struct sockaddr *)&serverAdress,
                                sizeof(serverAdress));
    if (connectStatus != 0) {
        cout << "Failed to connect";
        return connectStatus;
    } else {
        cout << "CONNECTED!" << endl;
    }

    // Perform SSL handshake
    if (SSL_connect(ssl) != 1) {
        cout << "SSL Handshake error" << endl;
    }

    while (true) {

        // Create a new message object
        Message message;

        // FIXME Seems to send multiple messages if content contains spaces
        // Get the content of the message from the user
        cout << "Type message to server: ";
        cin >> message.content;

        // Exit the program if the user typed "exit" as the content of the
        // message
        if (message.content == "exit") {
            cout << "BYE!" << endl;
            break;
        }

        cout << endl;
        message.username = "Client 1";

        // Send the content of the message to the server with encryption
        // .data() gets a pointer to the content of the message
        SSL_write(ssl, message.toString().data(),
                  strlen(message.toString().data()));
    }

    // Clean up
    // Deallocate the SSL objects
    SSL_free(ssl);

    // Deallocate the SSL context
    SSL_CTX_free(ctx);

    // Deallocate the SSL error strings
    ERR_free_strings();

    // Close the client socket
    close(clientSocket);

    return 0;
}
